#!/usr/bin/perl -w

# dhcp leases to djbdns data file script. 
# written september 2001, by Em Tonkin
# Windows dhcp information provided by Jason Grider, of fiskars.com
# (which is apparently THE place for resin faux ceramics, ladies and gentlemen - 
# support your local branch today!)
# This program is GPL'd. Even if it blows up your network and murders your mother,
# you don't go blaming me.

# Why this is an adaption of various things off the net rather than just being the one
# available from www.thismetalsky.org which OTOH had a lot to do with it:
# We have a different network which has both static machines, non-static reserved dhcp,
# and truly dynamic dhcp for random people and machines. This means we can't just go 
# rewriting the entire data file, because our mail server will go 'way and never come 
# back and so on. So we needed to be a little bit more clever, kind of thing, in order to 
# avoid utterly messing up the dns...  

# ____________________________________________________________________________________
# EXAMPLE CONFIG FILE- LEAVE HERE FOR INFORMATION PURPOSES...
# Caps or nocaps doesn't matter; I'll make the whole thing lowercase anyway, except for the inputfile and 
# output file... better be more careful with those

#OSTYPE=WINDOWS
#OSTYPE=DHCP

#OUTPUTFILE=/tmp/data_output

#INPUTFILE=./example_dhcp.lease
#INPUTFILE=./dhcpd.lease

#OUTPUT=CANONICAL
# OUTPUT=alias

#RESULTTYPE=REPLACE-FILE
# RESULTTYPE=APPEND-FILE
# RESULTTYPE=MERGE-FILES
# RESULTTYPE=STDOUT

# if it appends it replaces everything after a comment, unless the comment is not there
# At which point it assumes it's the first ever time its been run, adds a comment to the bottom of 
# the file, and then replaces everything after it...

# If it's supposed to merge, it performs possibly the most dangerous option in the sense that I might
# conceivably have left a bug in the function (it's more complex than just printing out/adding to the 
# bottom of the page, which is easy). This function is to open the current file, read it all out into the 
# same array that the originals are stored in, and then wander through the array making a final array from the
# current one, leaving no duplicates whatsoever, nor ones with blank/nonsense/NULL hostnames, one hopes.

# Eventually also will need a bunch of stuff in this file for telnet communication yes/no, 
# for placing things from NT onto a remote server, for example
# Also will need some other stuff - for restarting tinydns, mainly ie rerun make yes/no, and 
# whether to do it on this machine or a non-local machine (the things NT forces us to) 

# END OF EXAMPLE CONFIG FILE_______________________________________________________

# You can keep this empty for now, it'll fill itself up from the config file anyway
my %VARIABLES;

# This tells the system which variables are important. It's necessary for you to keep 
# this just as it is unless you intend to edit the script and make more required variables,
# in which case you can of course add them to the mix... 

my @NECESSARY_VARS=("OSTYPE","OUTPUTFILE","INPUTFILE","OUTPUT","RESULTTYPE");

# Initial variable definitions
if(scalar(@ARGV)>1){
	# There's apparently a command line thing going, so use those 

	foreach $element (@ARGV){
		@tmp=split("=",$element);
		$VARIABLES{$tmp[0]}=$tmp[1];
	}
 
} elsif($ARGV[0] eq "--help"){
	print "Help for $0\n\n";

	print "Either specify a file on the command line ie $0 /config/file/location\n";
	print "or give command line vars\n";


	exit;
} else {
	# Read all the vars from a config file...
	
	print "Assuming that you do not intend to use command line variables  \
	(use $0 --help for command line help info)\n";

	print "Checking for the existence of $ARGV[0]... \n";
	open (FILEHANDLE,"<$ARGV[0]") or die "File $ARGV[0] not found. Maybe you're confused; \n type \"--help\" for instructions\n";
	print "Using file $ARGV[0] \n";

	while(<FILEHANDLE>){
		# Let's do it this way (easier to edit than grep and map)
		# Which would otherwise be @tmp=grep {$_ !=~ /^\#/i || $_ eq "\n" } @INPUTFILE		
		# if the line I'm reading from the file starts with a # it's a comment. Or if it is only a \n it's empty
		# The below tests for these conditions and ignores them

		if($_ =~ /\#/i || $_ eq "\n"){

			# discard this line from the config file (do nothing - it's useless)

		} else { 
				# This is a useful variable; add it to the hash ;-)

			@tmp= split(/=/,$_);

			# little regexp to rid ourselves of that DAMN newline... chop and chomp DON'T
			$tmp[1]=~s/\n//i;
			$VARIABLES{$tmp[0]}=$tmp[1];

			#print " $tmp[0] and $tmp[1]\n";
		}
	}
	close (FILEHANDLE);	
	
}

# Structure of command line to call this script
# Required switches: location of the leases file, type of it (Win/dhcpd)
# Option to define how they should be written out - as a canonical definition (=foo.bar.com)
# or as a non-canonical alias -requires a canonical hostname elsewhere (+foo.bar.com)

# Let's just check that we got all the right variables there!
print join (",",keys( %VARIABLES )),"\n";

foreach $element (@NECESSARY_VARS) {
	if(!exists($VARIABLES{$element})){
		print "Necessary variable $element not defined \n";
		exit;
	}	
	#print "$VARIABLES{$element}";
}



# Make a DNS parser - you're going to need one to open and read that file out into convenient form for you...


$parser=DNSparser->new;


# Add standard lines...
foreach $element (keys(%VARIABLES)) {
        if ($element =~ /^KEEP/){
                $parser->add_standardline($VARIABLES{$element});
                print "Added line: $element $VARIABLES{$element}";
        }
}


# Give the parser the input filename
$parser->set_filename($VARIABLES{INPUTFILE});

# Just make sure that OSTYPE is in capitals, and therefore will be recognised even if somebody wrote it lowercase.

$VARIABLES{OSTYPE}=~ tr/a-z/A-Z/;


# If windows style file, parse as Windows
if($VARIABLES{OSTYPE} eq "WINDOWS"){
	$parser->parse_WINdhcp();
} else {
	# otherwise, parse as Linux/UNIX style DHCPD format file
	print "Parsing as DHCP format file...\n";
	$parser->parse_dhcp();
}

$VARIABLES{OUTPUT}=~ tr/a-z/A-Z/;
$VARIABLES{RESULTTYPE}=~ tr/a-z/A-Z/;

# Reminder for the programmer: 
# RESULTTYPE=REPLACE-FILE
# RESULTTYPE=APPEND-FILE
# RESULTTYPE=MERGE-FILES
# RESULTTYPE=STDOUT

if($VARIABLES{RESULTTYPE} eq "STDOUT"){

 	$parser->print_datafile_to_stdout($VARIABLES{OUTPUT});

} elsif ($VARIABLES{RESULTTYPE} eq "REPLACE-FILE") {

	$parser->print_as_new_datafile($VARIABLES{OUTPUTFILE},$VARIABLES{OUTPUT});

} elsif ($VARIABLES{RESULTTYPE} eq "APPEND-FILE") {

	$parser->print_as_appended_datafile($VARIABLES{OUTPUTFILE}, $VARIABLES{OUTPUT});

} elsif ($VARIABLES{RESULTTYPE} eq "MERGE-FILES") {

	#$parser->print_as_appended_datafile($VARIABLES{OUTPUTFILE}, $VARIABLES{OUTPUT});


	# There, we have them all in one file: Now, let's consider them
	# What we want to do is allow only one machine with each hostname 
	# Note, if you have a situation in which you WANT several IPs
	# to share one hostname ie. load balancing, I suggest you do not 
	# do this using MERGE-FILES. This algorithm is here merely in 
	# order to ensure that a basic dhcp dynamic dns system does not
	# cause confusion by having two machines with one hostname, 
	# and that if the same IP turns up with two different hostnames,
	# that the top one in the list is set to = and the one underneath
	# is set to +... this makes sense for our implementation since we 
	# give each dynamic IP a canonical hostname above the # comment,
	# and give it an alias if supplied by the holder of the lease after.

	$parser->clean_merge_file($VARIABLES{OUTPUTFILE},$VARIABLES{OUTPUT});
	
	# Note, this thing ain't finished yet :P
	# Update: Actually, you know, it is finished ;-)
	# I am quite proud of this one though you are warned that it doesn't
	# search for every one, only the more amazingly scary duplicates
	# and a couple of situations (same hostname, same IP)
	# somebody please give me some reader feedback on what situations they
	# want it to look for, and I'll update accordingly	
}

#Uncommenting this next line makes it run the update shell script to update tinydns's data file by running make...
#	$parser->make_data();




# Obselete entries. I should leave these alone if I were you ;-)
# $parser->return_dns_entries_for_debug();
# $parser->test_dns_package();
# ------------------------------------------------------------------------------

# End of program, start of various package definitions...












# beyond this, the namespace is all in DNSparser 








package DNSparser;

# This is the package everything happens from...
# It probably ought to return an array of DNSpackage objects...
sub new
{
	my $class=shift;

	# Let's define ourselves a big hash so our class can have a hash table of its very own...
	my $self={};
        bless($self,$class);
	#print "Created a new DNSparser";
        return $self;

}

sub make_data
{
	$self=shift;
	$result=exec("./tinydns_update.sh");
}
sub how_many_entries
{
	$self=shift;
        $number=$#{ $self->{IP}};
	#print "There are lots ($number) of entries...  \n";
	return $number;
}

sub set_hostname
{
        $self=shift;
        $hostname=shift;
        $number=$#{ $self->{IP} };
        if($hostname eq ""){
                print  "        $hostname\n";
                $hostname="NULL";
        }
        push @{ $self->{HOSTNAME} }, $hostname;
        $self->{HOSTNAME}[$number]=$hostname;
        print "Hostname set to $self->{HOSTNAME}[$#{ $self->{HOSTNAME} }]\n";
}

sub get_hostname
{
        $self=shift;
        $recallme=shift;
	#print "FROM GET_HOSTNAME *** Recallnumber is  $recallme ***\n";
        return $self->{HOSTNAME}[$recallme];
}

sub add_standardline 
{
	$self=shift;
	$standardline=shift;	
	push @{ $self->{STANDARDLINES}}, $standardline;
	print "Standard line added: $standardline \n";

}

sub set_IP
{
        $self=shift;
        $IP=shift;
        #print "IP is $IP and current size of system is ",how_many_entries(),"\n";
        push @{ $self->{IP} }, $IP;
        $number=$#{ $self->{IP}};
        $self->{IP}[$number]=$IP;
        #print "IP $number ($IP) set to $self->{IP}[$number]\n";
}

sub get_IP
{
        $self=shift;
        $recallnumber=shift;
        #print "$recallnumber is recallnumber and size of the database is $#{ $self->{IP}}\n";
        return $self->{IP}[$recallnumber];
}


sub set_filename
{
	$self=shift;
	$self->{FILENAME}=shift;	
	print "Using file $self->{FILENAME}\n";
}

sub clean_merge_file
{
	$self=shift;
	$Filename=shift;
	$Filetype=shift;
	print "Cleaning file $Filename...\n";
	
	# First, read the entire thing back into memory. 
	# If it doesn't start with an = or a +, ignore it entirely, it doesn't matter damn it.
	# Nb. this algorithm is horrifically simple ;-)
	# And utterly broken. We need to think about this for a few minutes I think...

	open (FILEHANDLE,"<$Filename") or die ("Could not open file $Filename for cleaning...\n");
	$count=0;
	$comment_number=undef;
	
	while(defined($line=<FILEHANDLE>)){
                if ($line =~ /^\#/) {
                        $comment_number=$count;
                }

		if(!$comment_number){
               		$tmparray[$count]=$line;
			print "$line";
			
		}
		$count++;
	}
	close FILEHANDLE;

	print "$comment_number\n";
	$count=0;	
	#@foreach $element (@tmparray) {
	#	unless ($element eq "\n"){
	#	@tmplinearray=split(':',$element);

		# We now have the elements of the thing, stuck in here and available to play with...
		# Check whether the hostname is included elsewhere, then check whether both entries are canonical
		# If they both are then complain
		# Check whether there is the same IP with two different hostnames, both canonical. 
		# If this happens, change one of them to an alias (preferably the one from the DHCP ;-)
		# Also add a complaint...
		# let's start with 'if it starts with an = or a +, otherwise ignore' since anything
		# that doesn't start with = or + is a different kind of entry, unless it starts with 
		# - at which point it's disabled anyway 

	#	print "$tmplinearray[0] : $tmplinearray[1] : $tmplinearray[2] \n";
		
		
	#	$count++;
	#	}
	#}

        %initial_char=(
                'CANONICAL','=',
                'ALIAS','+',
                      );

        open (FILEHANDLE1, ">$Filename") or die ("Could not open new data file ($Filename)...\n");


	# Try a grep version...
	$number_of_elements=$self->how_many_entries();
	foreach (@tmparray) {
		
		print FILEHANDLE1 $_;
	}
	print FILEHANDLE1 "# This is a Dynamic DNS comment \n";
	OUTER: for($i=0; $i<$number_of_elements; $i++){
		$hostname=$self->get_hostname($i);
		$IP=$self->get_IP($i);
		#If all is fine we'd put this entry in...
                $entry="$initial_char{$Filetype}$IP:$hostname:86400";

		# print "We've executed this loop $i times now with $entry \n";
		# @tmpduplicate=grep {$_ =~ /$hostname/} @tmparray;
		# Remember, tmparray are the ones that are in the file and MUSTN'T be changed cos mister
		# system administrator knew what he was doing when he put them in there...
		$ADD_DUP=false;
		INNER: foreach (@tmparray) {
			if ($_ =~ /$entry/){
				print "\tUtterly duplicate entry $entry : $_ \t- discarded\n";
				next OUTER;
			}
			@tmparrayinside=split(":",$_);

			#print "$#tmparrayinside $hostname \n";
			if($#tmparrayinside>0){
				if ($tmparrayinside[1] eq $hostname) {
					$duplicate=$_;
					print "\tDuplicate hostname?  $hostname and $tmparrayinside[1]";
					print "\tMachine IPs: $IP and $tmparrayinside[0]\n";
					if($_ =~ /^=/ && $Filetype eq "CANONICAL"){
					print "\tAnd they're both canonical, too!\n";
					print "\tLet's print a warning to messages and discard it...\n";
					print "\t(but do nothing else... what can we do?)\n";
					next OUTER;
					}	
				} 
			}
			if ($_ =~ /$IP/){
				$duplicate=$_;
				print "	Duplicate IP found: Hostnames: $hostname and $tmparrayinside[1] on  $IP $duplicate";
				if ($duplicate =~ /^=/ && $Filetype eq "CANONICAL") {
					print " \tChanging definition to alias for $entry \n";
					$entry =~ s/^=//;
					$ADD_DUP=true;
				}
			}
		}
		if($ADD_DUP eq true){
			$entry="+$entry";
		}
		print FILEHANDLE1 "$entry\n";
	}
	close FILEHANDLE;
}

sub print_as_appended_datafile
{
	$self=shift;
	$OutFilename=shift;
	print "OutFilename is: $OutFilename \n";
	$Filetype=shift;

	open (FILEHANDLE, "<$OutFilename") or die ("Could not open new data file ($OutFilename)...\n");
	
	# Note: If you have a really BIG dns system, you really don't want to 
	# use this particular option because I have to tell you that it'll use a lot 
	# of memory to temporarily buffer the entire contents of your current system...
	# Thousands of lines won't hurt, though.

	$tmpstring=0;
	$COMMENTREACHED=FALSE;
	while(defined($line=<FILEHANDLE>)){
		if($line =~ /^\#/){
			$COMMENTREACHED=TRUE;
			print "$line";
		}
		if($COMMENTREACHED eq FALSE){	
			$tmparray[$tmpstring]=$line;
		}else {
			print "COMMENTREACHED=true: discarding line $line \n";
		}
		$tmpstring++;
	}
	close FILEHANDLE;
	open (FILEHANDLE, ">$OutFilename") or die ("Could not open data file $OutFilename \n");

        %initial_char=(
                'CANONICAL','=',
                'ALIAS','+',
                      );

	foreach $element (@tmparray){
		# hopefully still has the \n attached...
		print "There's a line to print \n";
		print FILEHANDLE "$element";
	}
	print FILEHANDLE "# Dynamic DNS comment line\n";
        foreach $element (@{ $self->{STANDARDLINES}}){
                print FILEHANDLE "$element\n";
        }
	$number_of_entries=$self->how_many_entries();

        for($i=0; $i<$number_of_entries; $i++){
                unless($self->get_hostname($i) eq "NULL"){
                        print FILEHANDLE $initial_char{$Filetype},"",$self->get_IP($i),":",$self->get_hostname($i),":86400\n";
                }
        }
        close FILEHANDLE;


}


sub print_as_new_datafile
{
	$self=shift;
	$OutFilename=shift;
	print "OutFilename is: $OutFilename \n";
	$Filetype=shift;

	# Possible Filetypes 
	# OUTPUT=CANONICAL
	# OUTPUT=alias

	open (FILEHANDLE, ">$OutFilename") or die ("Could not open new data file ($OutFilename)...\n");			
	%initial_char=(
		'CANONICAL','=',
		'ALIAS','+',
			);
	#print "Am in print_as_new_datafile\n--------------------------------\n";
	$number_of_entries=$self->how_many_entries();		
	#print "Number of entries turned out to be : $number_of_entries \n";

	foreach $element (@{ $self->{STANDARDLINES}}){
		print FILEHANDLE "$element\n";
	}

	# Note that eventually I'm going to have to get a proper appreciation of the more complex side of dhcp
	# But right now I don't really see the point of doing anything more complicated than the below, 
	# Unless it turns out to be really useful, of course

	# I'll call it a job for next week or after I've got that firewall script done, guys ;-)

        for($i=0; $i<$number_of_entries; $i++){
		unless($self->get_hostname($i) eq "NULL"){
               		print FILEHANDLE $initial_char{$Filetype},"",$self->get_IP($i),":",$self->get_hostname($i),":86400\n";
		}
        }
	close FILEHANDLE;	

}

sub print_datafile_to_stdout
{
        $self=shift;
        $Filetype=shift;

        # Possible Filetypes 
        # OUTPUT=CANONICAL
        # OUTPUT=alias


        %initial_char=(
                'CANONICAL','=',
                'ALIAS','+',
                        );
        $number_of_entries=$self->how_many_entries();

        foreach $element (@{ $self->{STANDARDLINES}}){
                print "$element\n";
        }

        for($i=0; $i<$number_of_entries; $i++){
                unless($self->get_hostname($i) eq "NULL"){
                        print $initial_char{$Filetype},"",$self->get_IP($i),":",$self->get_hostname($i),":86400\n";
                }
        }


}

sub debugprint_formatted_NAME_IP_array
{
	$self=shift;
	# Don't even THINK about moving these lines away from the left margin. It'd break the script...
	# If you fiddled and then got yourself a complaint about missing end brackets and no compilation,
	# This is very possibly the place where it comes from ;-)
        $ip=shift;
        $name=shift;

format STDOUT =
@<<<<<<<<<<<<<<<<<<<<@<<<<<<<<<<<<<<<<<<<<<
	$name,                $ip
.
	write;
}

# function to parse Windows file into the form of an array of DNSpackage objects

sub parse_WINdhcp
{
	open (FILEHANDLE,$self->{FILENAME}) or die "File $self->{FILENAME}  not found. \
	Maybe you're confused; \n type \"--help\" for instructions\n";
        while(<FILEHANDLE>){
		# If you find that your dhcp server doesn't give the same sort of daft comments 
		# as Jason's clearly does, then you may need to change the line below...
		# since microsoft wossname seems to write all the actual information with numbers
		# at the beginning, you might want to try something like...
		# (and note, this is from imagination, I haven't tried it yet)
		# if ($_ =~ /^[a-zA-Z]/) don't use the line...
		# ie if the first character is alphanumerical....
		# or even better, if $_ =~ /^[0-9]/, use the line! ;-)

                if($_ =~ /^DHCP/i || $_  =~ /^Command/i){

                        # discard this line from the config file (do nothing - it's useless)

                } else {
                                # This is a useful variable; add it to the hash ;-)

                       # @tmp= split(/=/,$_);
                       # $VARIABLES{$tmp[0]}="$tmp[1]";
		
			# Graaaaaaaaah bloody windows can't do anything right...
			# When it sees a null hostname it reasons it has to put SOMETHING...
			# so it cheats by putting the resolutely bizarre and DNS-breaking (null)
			# Gah.
	
			#print $_," ";
			@tmp=split(" ",$_);
                        #print " $tmp[2]  $tmp[1]\n";
			#debugprint_formatted_NAME_IP_array(@tmp);

			# Let's fix that (null) ...
			$tmp[2] =~ s/\(null\)/NULL/g;

			$self->set_hostname($tmp[2]);
			$self->set_IP($tmp[1]);			
	                }
		}	
        close (FILEHANDLE);

		
}

# Function to parse Linux dhcpd leases file into the form of an array of DNSpackage objects

sub parse_dhcp
{
        open (FILEHANDLE,$self->{FILENAME}) or die "File $self->{FILENAME}  not found. \
        Maybe you're confused; \n type \"--help\" for instructions\n";


	print "Am in parse dhcp\n";
	# To help us work out what level of brackets we're on, I have a $level var...
	my $level=0;
	my $number_of_entries=0;

        while(<FILEHANDLE>){

		# Parse Linux's DHCP file (or whoever's lease file it is, provided it's in STANDARD format... not microsoft then)
		# Mind you, it will not appreciate malformed files. no sneakily forgetting closing brackets, now...	
	
        $_ =~ s/\"//g;
        $_ =~ s/\;//g;
        $_ =~ s/\,//g;

	# print $_;
	if ($_ =~ /\}/i)
	{
		$level=0;

		# increment the number of entries. By the by, there will be a number of entries with no hostname...
		# don't worry about it, we'll root them out later
	
                if (!$self->get_hostname($number_of_entries)) {
                        $self->set_hostname("NULL");
                        print "Was null \n";
                }
		$number_of_entries++;		
		print "Incremented number of entries : $number_of_entries \n ";
		print "-------------------------------- Got to closing \n";
	}

	if ($level>0){
		# We know the hostname's that little sucker sitting right there after the word 'client-hostname'...
		if($_ =~ m/client-hostname(.*)/i)
		{
			my $hostname=$1;
			$hostname =~ s/ //g;
			$self->set_hostname($hostname);
			print "Hostname  : $hostname \n";
		}
	}


	if ($_ =~ /\{/i)
	{
		#print $_;
		# This is the initial line of one... it has the opening bracket...
		# let's extract the IP addresses and put them in @DNS_ENTRIES...

		# oh, and since we have that bracket we increment the level... let's go up in the world!...
		$level=1;

		# Let's use a regular expression to get that IP...
		$_ =~ m/lease(.*)\{/i;
		$ip=$1;

		# Get rid of the spaces....
		$ip =~ s/ //g;

		print "***$ip*** \# $number_of_entries *********\n";
	
		$self->set_IP($ip);
	}

	}
        close (FILEHANDLE);

	print "Finally at the end; we have $number_of_entries or ",$self->how_many_entries()," entries\n";	
	
	for($i=0; $i<$number_of_entries; $i++){
		print " From same function : *** ",$self->get_IP($i)," $i: ",$self->get_hostname($i),"\n ";
	}
	


}




#sub test_dns_package
#{
#	my @arraytest;
#	$arraytest[0]=DNSpackage->new;
#	$arraytest[0]->set_hostname("bob");
#
#	#	$testDNS=DNSpackage->new;
#	#	$testDNS->set_hostname("Bob");
#
#	print "hostname is ",$arraytest[0]->get_hostname(),"\n";
#	#print "hostname is ",$testDNS->get_hostname(),"\n";
#}


# Function to write those out as part of a data file for djbdns

# preferably, function to get the output of this onto a linux server (for evil nt users),
# and to rerun the 'make' automatically. Otherwise, we could do that the other way round...
# OR by web or wget. Probably better to do it the other way, in fact - but must discuss this 
# with jason...

# function to rerun tinydns's 'make' on a local server...
